from functools import lru_cache, _make_key
import functools
from typing import Callable, Dict, Hashable, Tuple
import json

from nekoite_be_core.types.interfaces import IStrSerializer

try:
    import redis
except ImportError:
    pass

try:
    from functools import cache  # type: ignore
except ImportError:

    def cache(user_function: Callable):  # type: ignore
        'Simple lightweight unbounded cache.  Sometimes called "memoize".'
        return lru_cache(maxsize=None)(user_function)


__all__ = ["redis_func_cache", "lru_cache", "cache"]


def _default_key_maker(
    func: Callable, args: Tuple[Hashable, ...], kwargs: Dict[str, Hashable], typed: bool
) -> str:
    return func.__name__ + str(hash(_make_key(args, kwargs, typed)))


def redis_func_cache(
    r: redis.Redis,
    prefix: str,
    *,
    typed: bool = False,
    key_maker: Callable[
        [Callable, Tuple[Hashable, ...], Dict[str, Hashable], bool], str
    ] = None,
    serializer: IStrSerializer = json,
    **redis_kwargs,
):
    """
    Currently the serializer used is json instead of pickle for security issues,
    but pickle can also be used by settiing the `serializer=pickle`.
    """
    import redis

    if not isinstance(r, redis.Redis):
        raise RuntimeError("r should be a redis instance")

    if not key_maker:
        key_maker = _default_key_maker

    def _decorator(func):
        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            key = prefix + key_maker(func, args, kwargs, typed)
            v = r.get(key)
            if v:
                return serializer.loads(v)
            v = func(*args, **kwargs)
            r.set(key, serializer.dumps(v), **redis_kwargs)
            return v

        return _wrapper

    return _decorator



from typing import Any, Callable, Type, Union

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Query, scoped_session, sessionmaker

from nekoite_be_core.types.interfaces import ICommitable as _Commitable

__all__ = ["make_commit_decorator", "ModelBase", "make_model_base"]


def make_commit_decorator(session: _Commitable):
    """
    Generates a decorator that does session.commit() after the function
    is finished.

    Usage:

    ```python
    dbcommit = make_commit_decorator(session)

    @dbcommit
    def add_student():
        s = Student(name="John")
        s.insert()
    ```
    """

    def dbcommit(func: Callable) -> Callable:
        def do_and_commit(*args, **kwargs):
            func(*args, **kwargs)
            session.commit()

        return do_and_commit

    return dbcommit


class ModelBase(declarative_base()):  # type: ignore
    __abstract__ = True

    session: scoped_session
    query: Query

    def save(self):
        self.session.commit()

    def insert(self):
        self.session.add(self)

    def __getattribute__(self, __name: str) -> Any:
        try:
            return object.__getattribute__(self, __name)
        except AttributeError:
            pass
        v = getattr(self.query, __name, None)
        if v is not None:
            return v
        raise AttributeError(__name)

    @classmethod
    def commit(cls):
        cls.session.commit()


def make_model_base(session: Union[scoped_session, sessionmaker]) -> Type[ModelBase]:
    s = scoped_session(session) if isinstance(session, sessionmaker) else session

    class __Base(ModelBase):
        session = s
        query = session.query_property()

    return __Base



from nekoite_be_core.view import Response


__all__ = ["BEError"]


class BEError(RuntimeError, Response):
    def __init__(self, code: int, message: str = None, data: str = None) -> None:
        Response.__init__(self, code, message, data)
        RuntimeError.__init__(self, message)



DESCRIPTION = "A backend core utility library."

with open("README.md", "r", encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()


CLASSIFIERS = """\
Development Status :: 3 - Alpha
Intended Audience :: Developers
License :: OSI Approved :: MIT License
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.7
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3.9
Programming Language :: Python :: 3.10
Programming Language :: Python :: 3 :: Only
Programming Language :: Python :: Implementation :: CPython
Topic :: Software Development
Typing :: Typed
Operating System :: OS Independent
"""

INSTALL_REQUIREMENTS = """\
sqlalchemy>=1.4
marshmallow>=3.14
"""


def setup_pkg():
    import setuptools

    metadata = dict(
        name="nekoite-be-core",
        version="0.1.1",
        author="Nekoite",
        author_email="nekoite@rebuild.moe",
        license="MIT",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        url="https://github.com/nekoite/nekoite-be-core",
        classifiers=[_f for _f in CLASSIFIERS.split("\n") if _f],
        python_requires=">=3.7",
        install_requires=[_f for _f in INSTALL_REQUIREMENTS.split("\n") if _f],
        package_dir={"": "src"},
        packages=setuptools.find_packages("src"),
    )

    setuptools.setup(**metadata)


if __name__ == "__main__":
    setup_pkg()
